#!/usr/bin/python3
# coding:utf-8

# Copyright (c) 2019 ISB CORPORATION <http://www.isb.co.jp/>
# All rights reserved.

import sys
import os
import signal
import time
import json
import isb_wsn_serial

from time import sleep
from isb_wsn_serial import WiSunSerial


################################################################################
# Wi-SUN Dual Stack
################################################################################

def main():
	"""Wi-SUNメイン処理
	"""
	# 設定をjsonから読み込む
	ttyDev, init_settings, password, panc_addr = load_config()

	# シリアルポートオープン
	wsn_ser = WiSunSerial(ttyDev)
	if(wsn_ser.isOpen() == False): return 1

	# ハードウェアリセット
	wsn_ser.cmd_com_reset()
	# MACアドレス取得
	wsn_ser.cmd_com_get_addr()
	# 初期設定
	wsn_ser.cmd_com_init(init_settings)

	# アクティブスキャンでPANコーディネータを検索
	while(True):
		if(panc_addr == bytes.fromhex("FFFFFFFFFFFFFFFF")):
			# 初期接続モード
			pairId = int.from_bytes("HAN_INIT".encode("ascii"), byteorder='big')
		else:
			# 通常接続
			pairId = int.from_bytes(panc_addr, byteorder='big')

		# 全チャネルをアクティブスキャン
		for i in range(3):
			ret, scan_results = wsn_ser.cmd_com_activescan(6, 0x0003FFF0, True, pairId)
			if(len(scan_results) != 0):
				break
			else:
				sleep(1)

		if(len(scan_results) > 0):
			# 発見したPANコーディネータのチャネルとアドレスを設定
			init_settings[2] = scan_results[0][0]
			panc_addr = scan_results[0][1]
			wsn_ser.cmd_com_init(init_settings)
			save_config(init_settings=init_settings, panc_addr=panc_addr)
			break
		else:
			# 指定されたPANコーディネータが見つからなかったため、初期接続モードで接続を試みる
			panc_addr = bytes.fromhex("FFFFFFFFFFFFFFFF")
			continue


	# HAN動作開始
	while(True):
		ret = wsn_ser.cmd_han_start( 0x0000, panc_addr )
		if( ret == 0):
			break

	# HAN PANA認証情報設定
	wsn_ser.cmd_han_pana_setting( None, password)
	# HAN PANA開始
	wsn_ser.cmd_han_pana_start()
	# 親機とのPANA認証完了待ち合わせ
	wsn_ser.wait_for_pana_auth()
	# UDPポートオープン
	ret = wsn_ser.cmd_com_open_port(1111)
	ret = wsn_ser.cmd_com_open_port(2222)

	# 以降、受信ループ
	while True:
		recvData = wsn_ser.recv()
		cmdCode = isb_wsn_serial.get_cmd_code(recvData)
		cmdData = isb_wsn_serial.get_cmd_data(recvData)
		
		if(cmdCode == 0x6018): # データ受信通知
			srcIpv6 = cmdData[0:16]
			strSrcIpv6 = "".join(format(x, "02X")for x in srcIpv6)
			destPort = int.from_bytes(cmdData[18:20], byteorder='big')
			rssi = int.from_bytes([ cmdData[24] ], 'big', signed=True)
			payload = cmdData[27:]
			print("Received from {:s}, RSSI={:d}, Port={:d}".format(
				strSrcIpv6, rssi, destPort))
			print(" Data={:s}".format(
				"".join(format(x, "02X")for x in payload)))

			# send reply(same payload)
			ret = wsn_ser.cmd_com_send_data(
				strSrcIpv6, destPort, destPort, payload)

		else:
			continue


	return 0

################################################################################
# Application Config
################################################################################
_g_config = {}

def set_app_config(config_dict):
	global _g_config
	_g_config = config_dict.copy()
	return 

def get_app_config():
	global _g_config
	return _g_config.copy()

def load_config():
	"""設定をjsonファイルから読み込む
	"""
	f = open("ed_config.json", 'r')
	jsonObj = json.load(f)
	f.close()
	# グローバル設定を更新
	set_app_config(jsonObj)

	# シリアルポート
	ttyDev = jsonObj["SerialInterface"]["port"]
	# 動作モード
	mode = jsonObj["InitialSetting"]["actionMode"]
	# HAN Sleep機能
	if( jsonObj["InitialSetting"]["sleepSupport"] ):
		sleepSupport = 0x01
	else:
		sleepSupport = 0x00
	# チャネル
	channel = jsonObj["InitialSetting"]["channel"]
	# 送信出力
	txPower = jsonObj["InitialSetting"]["txpower"]

	# 初期設定
	init_settings = [
		mode,			# 動作モード(0x03:エンドデバイス)
		sleepSupport,	# HAN Sleep 機能設定(0x00:無効,0x01:有効)
		channel,		# チャネル(0x04～0x11)
		txPower			# 送信電力(0x00:20mW,0x01:10mW,0x02:1mW)
	]

	# パスワード
	password = 	jsonObj["HAN"]["password"].encode("ascii")
	# 接続先のPANコーディネータ
	panc_addr = bytes.fromhex(jsonObj["HAN"]["pancAddr"])

	return ttyDev, init_settings, password, panc_addr


def save_config(ttyDev=None, init_settings=None, 
	 password=None, panc_addr=None ):
	"""設定をjsonファイルに保存する
	"""
	config = get_app_config()

	if(ttyDev != None):
		# シリアルポート
		config["SerialInterface"]["port"] = ttyDev

	# 初期設定
	if(init_settings != None):
		config["InitialSetting"]["actionMode"] = init_settings[0]
		# HAN Sleep機能
		config["InitialSetting"]["sleepSupport"] = (init_settings[1] == 0x01)
		# チャネル
		config["InitialSetting"]["channel"] = init_settings[2]
		# 送信出力
		config["InitialSetting"]["txpower"] = init_settings[3]


	# パスワード
	if(password != None):
		config["HAN"]["password"] = password.decode('ascii')

	# 接続先のPANコーディネータ
	if(panc_addr != None):
		config["HAN"]["pancAddr"] = "".join(format(x, "02X")for x in panc_addr)

	try:
		# グローバル設定を更新
		set_app_config(config)
		# ファイル書込み
		basedir = os.path.dirname(os.path.abspath(__file__))
		f = open(basedir+"/ed_config.json", 'w')
		json.dump(config, f, indent=4)
		f.close()
	except:
		print('save_config() Error')
		return 1

	return 0


################################################################################

def sigint_handler(signal, frame):
	print('\nterminated by user')
	sys.exit(0)

if __name__ == '__main__':
	signal.signal(signal.SIGINT, sigint_handler)
	ret = main()
	sys.exit(ret)
