#!/usr/bin/env python3
# coding:utf-8

# Copyright 2019 ISB Corp. <http://www.isb.co.jp/>
# All rights reserved.

import sys
import serial
import binascii
import datetime
from serial.tools import list_ports
from time import sleep
import time

VERSION = "1.3.3"

################################################################################
# defines
################################################################################
UNIQUECODE_REQUEST  = [0xD0, 0xEA, 0x83, 0xFC]
UNIQUECODE_RESPONSE = [0xD0, 0xF9, 0xEE, 0x5D]
UNIQUECODE_NOTIFY   = [0xD0, 0xF9, 0xEE, 0x5D]

# 動作モード
HAN_MODE_PANC = 0x01
HAN_MODE_CRDI = 0x02
HAN_MODE_DEV  = 0x03
HAN_MODE_DUAL = 0x05
# 送信出力
TX_POWER_20MW = 0x00
TX_POWER_10MW = 0x01
TX_POWER_1MW  = 0x02

COMMAND_SUCCESS = 0x01

# データ送信 詳細結果
UDP_SEND_SUCCESS          = 0x00
UDP_SEND_FAIL_360S_LIMIT  = 0x02
UDP_SEND_FAIL_CCA         = 0x03
UDP_SEND_FAIL_NACK        = 0x05
UDP_SEND_FAIL_OTHER       = 0x08
UDP_SEND_QUEUING          = 0x0F
UDP_SEND_NOT_QUEUING      = 0x00
UDP_SEND_QUEUING_SUCCESS  = 0x10
UDP_SEND_QUEUING_FAIL     = 0x20

# インダイレクト送信成功
UDP_SEND_ID_SUCCESS       = 0x1F
UDP_SEND_ID_MULTI_SUCCESS = 0x10

################################################################################
# for debug
################################################################################
EnableDebugSerial = True
"""シリアル通信のデバッグ有効フラグ
True: シリアル通信ログを有効にする
"""

################################################################################

g_recv_buf = [] # 受信データ格納バッファ
def recv_cmd_advanced(serial_obj):
	"""UART IFコマンド受信
	"""
	cmd_buf = []
	gb_buf = []
	global g_recv_buf
	error_flg = False

	# ヘッダ部受信
	while(True):
		if( len(g_recv_buf) < 4):
			read_buf = serial_obj.read(1)
			if( len(read_buf) == 0):
				# 受信タイムアウト
				error_flg = True
				break
			else:
				g_recv_buf += read_buf
				continue

		# ユニークコード検索
		idxUni = 0
		while( idxUni < len(UNIQUECODE_RESPONSE) ):
			if( len(g_recv_buf) < len(UNIQUECODE_RESPONSE) ):
				break

			if( g_recv_buf[idxUni] != UNIQUECODE_RESPONSE[idxUni]):
				gb_buf += g_recv_buf[:idxUni+1]
				g_recv_buf = g_recv_buf[idxUni+1:]
				idxUni = 0
			else:
				idxUni += 1

		# ユニークコードが一致していないため、後続のデータを待ち受ける
		if( idxUni != len(UNIQUECODE_RESPONSE) ):
			continue

		# 残りのヘッダ部受信
		if( len(g_recv_buf) < 12):
			g_recv_buf += serial_obj.read( 12 - len(g_recv_buf) )

		break # end of loop

	# ヘッダ部を受信しているか？
	if( len(g_recv_buf) < 12 ) or ( error_flg ):
		#print("recv_cmd:[err] timeout. header(len={:d})".format(len(g_recv_buf)))
		gb_buf += g_recv_buf[:]
		cmd_buf = []
		error_flg = True
	else:
		# reception of the header area was complete.
		cmd_buf = g_recv_buf[:12+1]

	# ヘッダ部 チェックサム検証
	if( not error_flg ):
		head_chksum = 0
		for byte in g_recv_buf[:8]:
			head_chksum += byte

		recv_chksum = int.from_bytes(g_recv_buf[8:10], byteorder='big')
		if(head_chksum != recv_chksum):
			print("recv_cmd:[err] header checksum err. {:04X}!={:04X}." \
				.format(head_chksum, recv_chksum))
			g_recv_buf = g_recv_buf[12:]
			error_flg = True

	# データ部受信
	if( not error_flg ):
		body_len = int.from_bytes(cmd_buf[6:8], byteorder='big') - 4
		total_cmd_len = 12 + body_len
		if( len(g_recv_buf) >= total_cmd_len ):
			# 受信済み
			cmd_buf = g_recv_buf[:total_cmd_len+1]
		else:
			# データ部受信
			req_len = total_cmd_len - len(cmd_buf)

			while(True):
				follow_data = serial_obj.read( req_len )
				g_recv_buf += follow_data

				if( len(follow_data) == 0) or (len(follow_data) == req_len):
					# タイムアウト、または成功
					break
				else:
					# 指定サイズより少ない
					req_len -= len(follow_data)
					continue
			# 受信バッファ全てをコマンドバッファにコピー(上書き)
			cmd_buf = g_recv_buf[:]

		if(len(cmd_buf) < total_cmd_len):
			print("recv_cmd:[err] timeout. body(buf={:d},total={:d})" \
				.format(len(cmd_buf),total_cmd_len))
			g_recv_buf = g_recv_buf[12:]
			error_flg = True

	# データ部 チェックサム検証
	if( not error_flg ):
		body_chksum = 0
		for byte in cmd_buf[12:]:
			body_chksum += byte
		body_chksum &= 0xFFFF

		recv_chksum = int.from_bytes(cmd_buf[10:12], byteorder='big')
		if body_chksum != recv_chksum:
			print("recv_cmd:[err] body checksum err. {:04X}!={:04X}" \
				.format(body_chksum, recv_chksum))
			g_recv_buf = g_recv_buf[12:]
			#error_flg = True
		else:
			# reception of the body area was complete.
			g_recv_buf = g_recv_buf[total_cmd_len:]

	# データ部チェックサム検証エラー以外はバッファクリア
	if( error_flg ):
		g_recv_buf = []

	if( len(gb_buf) > 0 ):
		print("recv_cmd:[info] prefix garbage={:s}".format(
			"".join(format(x, "02X")for x in gb_buf)) )

	return bytes(cmd_buf)

def send_cmd(serial_obj, data):
	"""UART IFコマンド送信
	"""
	ret = serial_obj.write(data)
	return 0 if(len(data) == ret) else 1

def build_cmd(uniqueCode, cmdCode, cmdData):
	"""UART IF コマンド作成
	"""
	command = uniqueCode[:]
	body = []
	headerCheckSum = 0
	bodyCheckSum = 0

	# コマンドコード
	command += cmdCode.to_bytes(2, byteorder='big')

	# データ部の長さ
	cmdLen = len(cmdData) + 4
	command += cmdLen.to_bytes(2, byteorder='big')

	# ヘッダ部のチェックサム
	for item in command:
		headerCheckSum += item
	headerCheckSum &= 0xFFFF
	command += headerCheckSum.to_bytes(2, byteorder='big')

	# データ部のチェックサム
	cmdLen = len(cmdData)
	if(cmdLen > 0):
		body = cmdData[:]
		for item in body:
			bodyCheckSum += item
	bodyCheckSum &= 0xFFFF
	command += bodyCheckSum.to_bytes(2, byteorder='big')

	command += body
	return bytes(command)

def build_req_cmd(cmdCode, cmdData = None):
	"""要求コマンド作成
	"""
	if cmdData == None:
		cmdData = []
	return build_cmd(UNIQUECODE_REQUEST, cmdCode, cmdData)

def build_rsp_cmd(cmdCode, cmdData = None):
	"""応答コマンド作成
	"""
	if cmdData == None:
		cmdData = []
	return build_cmd(UNIQUECODE_RESPONSE, cmdCode, cmdData)

def get_cmd_code(cmd_buf):
	return int.from_bytes(cmd_buf[4:6], byteorder='big')

def get_cmd_data(cmd_buf):
	return cmd_buf[12:]

def input_dest_mac(addr=None):
	"""宛先MACアドレス入力
	"""
	ret_addr = None
	input_str = addr

	if(addr==None):
		print('Please input destination MAC address(8byte).')
		print('\tex: 0123456789ABCDEF')

	while(ret_addr==None):
		if(input_str!=None)and(len(input_str) == 16):
			ret_addr = bytes.fromhex(input_str)
		else:
			input_str = input('>>> ')

	return ret_addr

def mac_from_ipv6(ipv6):
	"""EUI64 -> IPv6変換
	"""
	ifId = bytearray(ipv6[-8:])
	ifId[0] = ifId[0] ^ 0x02 # Universal/Local bit inversion
	strIfid = "".join(format(x, "02X")for x in ifId)
	eui64 = "%s" % (strIfid)
	return eui64

def mac2ipv6(eui64):
	"""EUI64 -> IPv6変換
	"""
	ifId = bytearray(eui64[:])
	ifId[0] = ifId[0] ^ 0x02 # Universal/Local bit inversion
	strIfid = "".join(format(x, "02X")for x in ifId)
	ipv6 = "FE80000000000000%s" % (strIfid)
	return ipv6

def input_pana_pass(pana_pass=None):
	"""PANAパスワード入力
	"""
	ret_pass = None
	input_str = pana_pass

	if(pana_pass==None):
		print('Please input PANA password(16byte).')
		print('\tex(ascii): 30313233343536373839414243444546')
		print('\tex(char): 0123456789ABCDEF')

	while(ret_pass == None):
		if(input_str!=None)and(len(input_str) == 32):
			ret_pass = bytes.fromhex(input_str)
		elif(input_str!=None)and(len(input_str) == 16):
			ret_pass = bytes([ord (s) for s in input_str])
		else:
			input_str = input('>>> ')

	return ret_pass


class WiSunSerial():
	"""serial communication class for Wi-SUN Dual Stack
	"""
	UART_BAUDRATE = 115200   # bps
	UART_TIMEOUT = 10.0      # second

	def __init__(self,_tty_dev, _rtscts=False):
		self.serialObj = serial.Serial()
		self.serialObj.port = _tty_dev
		self.serialObj.baudrate = self.UART_BAUDRATE
		self.serialObj.timeout = self.UART_TIMEOUT
		self.serialObj.rtscts = _rtscts
		try:
			self.serialObj.open()
		except:
			print("Port open error:{:s}".format(self.serialObj.port))
			print(" port list:")
			ports = list_ports.comports()	# ポートデータを取得
			for info in ports: print("\t",info)

		self.hanMode = HAN_MODE_PANC
		self.isSleepingDevice = False

	def __del__(self):
		self.serialObj.close()

	def isOpen(self):
		return self.serialObj.isOpen()

	def setTimeout(self, time):
		self.serialObj.timeout = time

	def __str__(self):
		str =  'class(WiSunSerial) v{0}\n'.format(VERSION)
		str += ' tty device  :{:s}\n'.format(self.serialObj.port)
		str += ' baudrate    :{:d}\n'.format(self.serialObj.baudrate)
		str += ' timeout     :{:f}\n'.format(self.serialObj.timeout)
		str += ' rtscts      :{0}\n'.format(self.serialObj.rtscts)
		return str

	def recv(self):
		recvData = recv_cmd_advanced(self.serialObj)
		now = datetime.datetime.now()
		if(len(recvData) == 0):
			return recvData

		if(EnableDebugSerial):
			strRecv = "".join(format(x, "02X")for x in recvData)
			strNow = now.strftime("%Y/%m/%d %H:%M:%S.") + "%03d" % (now.microsecond // 1000)
			print(" {:s} Rx: {:s} {:s} {:s} {:s} {:s} {:s}".format(
				strNow,strRecv[0:8],strRecv[8:12],strRecv[12:16],
				strRecv[16:20],strRecv[20:24],strRecv[24:]))

		return recvData

	def flush_recv_buf(self):
		self.setTimeout(0.1)
		while(True):
			recvData = self.recv()
			if(len(recvData)==0):
				self.setTimeout(self.UART_TIMEOUT)
				return 0

	def send(self,data):
		ret = send_cmd(self.serialObj,data)
		now = datetime.datetime.now()

		if(EnableDebugSerial):
			strSend = "".join(format(x, "02X")for x in data)
			strNow = now.strftime("%Y/%m/%d %H:%M:%S.") + "%03d" % (now.microsecond // 1000)
			print(" {:s} Tx: {:s} {:s} {:s} {:s} {:s} {:s}".format(
				strNow,strSend[0:8],strSend[8:12],strSend[12:16],
				strSend[16:20],strSend[20:24],strSend[24:]))

		return ret

	def chk_response(self, cmdDoc, rspCmd, rspRet=None):
		recvData = self.recv()
		cmpCmd = build_rsp_cmd(rspCmd, [rspRet])
		print(cmdDoc, end=":")
		if(recvData == cmpCmd):
			print("Success"); return 0
		else:
			print("Failure"); return 1

	def wait_for_recv_command(self, cmdCode):
		while(True):
			uartcmd = self.recv()
			code = get_cmd_code(uartcmd)
			data = get_cmd_data(uartcmd)
			if(code == cmdCode):
				break

		return data

	def cmd_com_get_addr(self):
		"""共通:MACアドレス取得(cmd=0x000E->0x200E)
		"""
		sendData = build_req_cmd(0x000E)
		self.send(sendData)
		# response
		recvData = self.recv()
		cmdData = get_cmd_data(recvData)
		if(get_cmd_code(recvData) == 0x200E):
			if(cmdData[0] == COMMAND_SUCCESS):
				print("MACアドレス取得: Success")
				strAddr = "".join(format(x, "02X")for x in cmdData[1:])
				print(" MAC address: {:s}".format(
	 				strAddr) )
				return 0

		print("MACアドレス取得: Failure"); return 1

	def cmd_com_get_ver(self):
		"""共通:バージョン情報取得(cmd=0x006B->0x206B)
		"""
		sendData = build_req_cmd(0x006B)
		self.send(sendData)
		# response
		recvData = self.recv()
		cmdData = get_cmd_data(recvData)
		if(get_cmd_code(recvData) == 0x206B):
			if(cmdData[0] == COMMAND_SUCCESS):
				print("バージョン情報取得: Success")
				fwid = int.from_bytes(cmdData[1:3], byteorder='big')
				rev = int.from_bytes(cmdData[5:9], byteorder='big')
				print(" Firmware Version:{:04X}-{:02X}-{:02X}-{:08X}({:d})".format(
	 				fwid, cmdData[3], cmdData[4], rev, rev ) )
				return 0

		print("バージョン情報取得: Failure"); return 1

	def cmd_com_reset(self):
		"""共通:ハードウェアリセット(cmd=0x00D9->0x6019)
		"""
		sendData = build_req_cmd(0x00D9)
		self.send(sendData)
		# response
		while(True):
			recvData = self.recv()
			cmdCode = get_cmd_code(recvData)
			if(cmdCode == 0x6019):
				break
			elif(cmdCode == 0x60DA):
				self.send(sendData)
			else:
				continue

		print("ハードウェアリセット: Success")
		return 0

	def cmd_com_init(self,cmd_data):
		"""共通:初期設定(cmd=0x005F->0x205F)
		"""
		sendData = build_req_cmd(0x005F, cmd_data)
		self.send(sendData)
		# response
		ret = self.chk_response("初期設定",0x205F, COMMAND_SUCCESS)
		if(ret == 0):
			self.hanMode = cmd_data[0]
			if(cmd_data[1] == 0x01) and (self.hanMode == HAN_MODE_DEV):
				self.isSleepingDevice = True

		return ret

	def cmd_han_start(self,panid,panc_mac=None):
		"""HAN:HAN動作開始(cmd=0x000A->0x200A)
		"""
		if(self.hanMode == HAN_MODE_PANC) or (self.hanMode == HAN_MODE_DUAL):
			cmd_data = panid.to_bytes(2, byteorder='big')
		elif( panc_mac == None ):	# Coordinator or End device
			cmd_data = [0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF]
		else:
			cmd_data = panc_mac

		sendData = build_req_cmd(0x000A, cmd_data)
		self.send(sendData)

		# response
		while(True):
			recvData = self.recv()
			cmdData = get_cmd_data(recvData)
			if(get_cmd_code(recvData) == 0x200A):
				print("HAN動作開始", end=":")
				result = cmdData[0]
				if(result == COMMAND_SUCCESS):
					channel = cmdData[1]
					panid = int.from_bytes(cmdData[2:4], byteorder='big')
					print("Success")
					strMac = "".join(format(x, "02X")for x in cmdData[4:12])
					if(len(strMac) > 0):
						rssi = cmdData[12] - 256
						print(" MAC Connected to \'{:s}\', rssi={:d}dBm".format(strMac,rssi))
					ret = 0; break
				else:
					print("Failure"); ret = 1; break
		return ret

	def cmd_han_end(self):
		"""HAN:HAN動作終了(cmd=0x000B->0x200B)
		"""
		sendData = build_req_cmd(0x000B)
		self.send(sendData)
		# response
		ret = self.chk_response("HAN 動作終了",0x200B, COMMAND_SUCCESS)
		return ret

	def cmd_rtb_start(self):
		"""Bルート:Bルート動作開始(cmd=0x0053->0x2053)
		"""
		sendData = build_req_cmd(0x0053)
		self.send(sendData)

		# response
		recvData = self.recv()
		cmdData = get_cmd_data(recvData)
		print("Bルート動作開始", end=":")
		if(get_cmd_code(recvData) == 0x2053):
			result = cmdData[0]
			if(result == COMMAND_SUCCESS):
				channel = cmdData[1]
				panid = int.from_bytes(cmdData[2:4], byteorder='big')
				print("Success")
				strMac = "".join(format(x, "02X")for x in cmdData[4:12])
				rssi = cmdData[12] - 256
				print(" MAC Connected to \'{:s}\', rssi={:d}dBm".format(strMac,rssi))

				return 0

		print("Failure"); return 1

	def cmd_han_switch_connect_mode(self,mode):
		"""HAN:HAN受け入れ接続モード切り替え(cmd=0x0025->0x2025)
		"""
		sendData = build_req_cmd(0x0025, [mode])
		self.send(sendData)
		# response
		ret = self.chk_response("HAN受け入れ接続モード切り替え",0x2025,
 				COMMAND_SUCCESS)
		return ret

	def cmd_han_pana_setting(self, mac, password):
		"""HAN:PANA認証情報設定(cmd=0x002C->0x202C)
		"""
		if(self.hanMode == HAN_MODE_PANC) or (self.hanMode == HAN_MODE_DUAL):
			cmd_data = mac[:]
			cmd_data += password[:]
		else:	# Coordinator or End device
			cmd_data = password

		sendData = build_req_cmd(0x002C, cmd_data)
		self.send(sendData)
		# response
		ret = self.chk_response("HAN PANA認証情報設定",0x202C,
 				COMMAND_SUCCESS)
		return ret

	def cmd_rtb_pana_setting(self, rtb_id, rtb_pass):
		"""Bルート:PANA認証情報設定(cmd=0x0054->0x2054)
		"""
		cmd_data = rtb_id[:]
		cmd_data += rtb_pass[:]
		sendData = build_req_cmd(0x0054, cmd_data)
		self.send(sendData)
		# response
		ret = self.chk_response("Bルート PANA認証情報設定",0x2054,
 				COMMAND_SUCCESS)
		return ret

	def cmd_han_get_group_key(self):
		sendData = build_req_cmd(0x0028)
		"""HAN:グループ鍵取得(cmd=0x0028->0x2028)
		"""
		self.send(sendData)
		# response
		recvData = self.recv()
		cmdData = get_cmd_data(recvData)
		if(get_cmd_code(recvData) == 0x2028):
			if(cmdData[0] == COMMAND_SUCCESS):
				print("HANグループ鍵取得: Success")
				groupKey = "".join(format(x, "02X")for x in cmdData[1:17])
				keyId = cmdData[17]
				print(" HAN Key: {:02X}, {:s}".format(keyId,groupKey) )
				return 0

		print("HANグループ鍵取得: Failure")
		return 1

	def cmd_han_pana_start(self):
		"""HAN:PANA開始(cmd=0x003A->0x203A)
		"""
		sendData = build_req_cmd(0x003A)
		self.send(sendData)
		# response
		recvData = self.recv()
		cmdData = get_cmd_data(recvData)
		print("HAN PANA開始", end=":")
		if(get_cmd_code(recvData) == 0x203A):
			if(cmdData[0] == COMMAND_SUCCESS):
				print("Success")
				if(self.hanMode != HAN_MODE_PANC) and (self.hanMode != HAN_MODE_DUAL):
					# Coo(dinator or End device
					strPancMac = "".join(format(x, "02X")for x in cmdData[1:])
					print(" PANA auth started with \'{:s}\'".format(strPancMac))

				return 0

		print("Failure")
		return 1

	def cmd_rtb_pana_start(self):
		"""Bルート:PANA開始(cmd=0x0056->0x2056)
		"""
		sendData = build_req_cmd(0x0056)
		self.send(sendData)
		# response
		ret = self.chk_response("Bルート PANA開始",0x2056, COMMAND_SUCCESS)
		return ret

	def cmd_ota_client_start(self):
		"""OTA:OTA Client開始(cmd=0x0201->0x2201)
		"""
		sendData = build_req_cmd(0x0201)
		self.send(sendData)
		# response
		ret = self.chk_response("OTA Client開始",0x2201,COMMAND_SUCCESS)
		return ret

	def cmd_ota_client_end(self):
		"""OTA:OTA Client終了(cmd=0x0201->0x2201)
		"""
		sendData = build_req_cmd(0x0202)
		self.send(sendData)
		# response
		ret = self.chk_response("OTA Client終了",0x2202,
 				COMMAND_SUCCESS)
		return ret

	def cmd_com_open_port(self,udp_port):
		"""共通:UDPポートOpen(cmd=0x0005->0x2005)
		"""
		cmd_data = udp_port.to_bytes(2, byteorder='big')
		sendData = build_req_cmd(0x0005, cmd_data)
		self.send(sendData)
		# response
		ret = self.chk_response("UDPポートOpen({:d})".format(udp_port),
				0x2005,COMMAND_SUCCESS)
		return ret

	def cmd_com_close_port(self,udp_port):
		"""共通:UDPポートClose(cmd=0x0006->0x2006)
		"""
		cmd_data = udp_port.to_bytes(2, byteorder='big')
		sendData = build_req_cmd(0x0006, cmd_data)
		self.send(sendData)
		# response
		ret = self.chk_response("UDPポートClose({:d})".format(udp_port),
				0x2006,COMMAND_SUCCESS)
		return ret

	def cmd_han_deepsleep(self):
		"""HAN:DeepSleep要求(cmd=0x00DA->0x20DA)
		"""
		self.do_cmd_han_deepsleep_req()
		return self.do_cmd_han_deepsleep_res()

	def do_cmd_han_deepsleep_req(self):
		sendData = build_req_cmd(0x00DA)
		self.send(sendData)
		return 0

	def do_cmd_han_deepsleep_res(self):
		# response
		ret = self.chk_response("DeepSleep要求",0x20DA,COMMAND_SUCCESS)
		return ret

	def cmd_han_wakeup(self):
		"""空コマンド送信->WakeUp通知確認(cmd=0x60DA)
		"""
		# 空コマンド送信
		self.send([0x00])

		# WakeUp通知受信
		recvData = self.recv()
		cmdData = get_cmd_data(recvData)
		if(get_cmd_code(recvData) == 0x60DA):
			print("WakeUp通知: Success")
			return 0
		else:
			print("WakeUp通知: Failure")
			return 1


	def cmd_com_send_data(self,ipv6,src_port,dest_port,dataArray):
		"""共通:データ送信(cmd=0x0008->0x2008->0x6037 or 0x6038)
		"""
		dataArrayLen = len(dataArray)
		reqSendData = "%s%04X%04X%04X" % (
			ipv6,
 			src_port,
 			dest_port,
			dataArrayLen
		)
		cmd_data = bytes.fromhex(reqSendData)
		cmd_data += dataArray
		sendData = build_req_cmd(0x0008, cmd_data)
		self.send(sendData)

		# response
		preTime = time.perf_counter()
		while(True):
			recvData = self.recv()
			cmdCode = get_cmd_code(recvData)
			cmdData = get_cmd_data(recvData)
			if(cmdCode == 0x2008):
				break

			deltaSec = time.perf_counter() - preTime
			serialTO = self.serialObj.timeout
			if(deltaSec > serialTO):
				print("データ送信:Failure, Response Timeout({:.1f}sec)".format(serialTO))
				return 3

		flg_failure = True
		cmdResp = cmdData[0]
		if(cmdCode == 0x2008) and (len(cmdData) >= 2):
			sendRslt = cmdData[1]
			part = cmdData[2:]
			if(cmdResp == COMMAND_SUCCESS):
				if( sendRslt == UDP_SEND_SUCCESS ):
					flg_failure = False
				elif(sendRslt == UDP_SEND_ID_SUCCESS ):
					flg_failure = False
				elif(sendRslt == UDP_SEND_ID_MULTI_SUCCESS ):
					flg_failure = False

		if( flg_failure ):
			print("データ送信:Failure", end=", ")
			if(len(cmdData) >= 2):
				print("cmdResp={:02X}, SendRslt={:02X}".format(cmdResp, sendRslt))
				# 失敗要因の表示
				self.print_result_detail_tx(sendRslt)
				self.print_result_detail_queuing(sendRslt)

				if( (sendRslt&0x0F) == UDP_SEND_FAIL_360S_LIMIT):
					return 2
			else:
				print("cmdResp={:02X}".format(cmdResp))

			return 1

		if(sendRslt != UDP_SEND_ID_SUCCESS):
			# ユニキャストのインダイレクト通信でない場合、送信成功している
			# マルチキャストの場合、何台送信するかわからないため送信通知待ち合わせをスキップする
			#print("データ送信:Success")
			return 0

		# 送信先がSleepingDeviceの場合、HAN インダイレクトキュー送信通知を待ち合わせる
		print("\tWaiting for send indirect result notification...")
		numOfFragment = 1
		while( numOfFragment >= 1 ):
			recvData = self.recv()
			cmdCode = get_cmd_code(recvData)
			cmdData = get_cmd_data(recvData)
			if(cmdCode == 0x6037):
				numOfFragment = cmdData[1]
				if( cmdData[0] == 0x00 ): break
				else:
					destIp = cmdData[2:18]
					strDestIp = "".join(format(x, "02X")for x in destIp)
					print("データ送信(Indirect):Failure(result={:02X}, dest={:s})" \
						.format(cmdData[0],strDestIp))
					# 失敗要因の表示
					self.print_result_detail_tx(cmdData[0])
			elif(cmdCode == 0x6036):
				print("データ送信(Indirect):Failure. del {:d} packets".format(cmdData[0]))
				return 1
			else:
				continue

		# 送信成功(Indirect)
		#print("データ送信(Indirect):Success")
		return 0

	def print_result_detail_tx( self, resltTx ):
		detail = resltTx&0x0F
		print("\tUDPデータ送信の詳細結果(Z)")
		if(detail == UDP_SEND_SUCCESS):
			print("\t\t0x{:02X}:成功".format(detail))
		elif(detail == UDP_SEND_FAIL_360S_LIMIT):
			print("\t\t0x{:02X}:送信総和制限のため送信NG".format(detail))
		elif(detail == UDP_SEND_FAIL_CCA):
			print("\t\t0x{:02X}:CCA Failedのため送信NG".format(detail))
		elif(detail == UDP_SEND_FAIL_NACK):
			print("\t\t0x{:02X}:ACK未受信のため送信NG".format(detail))
		elif(detail == UDP_SEND_FAIL_OTHER):
			print("\t\t0x{:02X}:その他失敗要因のため送信NG".format(detail))
		elif(detail == UDP_SEND_QUEUING):
			print("\t\t0x{:02X}:送信なし（キューイングのみ）".format(detail))
		else:
			print("\t\t0x{:02X}:unkown result".format(detail))

	def print_result_detail_queuing( self, resltTx ):
		detail = resltTx&0xF0
		print("\tインダイレクトキューイングの詳細結果(Y)")
		if(detail == UDP_SEND_NOT_QUEUING):
			print("\t\t0x{:02X}:インダイレクトキューにキューイングなし".format(detail))
		elif(detail == UDP_SEND_QUEUING_SUCCESS):
			print("\t\t0x{:02X}:インダイレクトキューにキューイングした".format(detail))
		elif(detail == UDP_SEND_QUEUING_FAIL):
			print("\t\t0x{:02X}:インダイレクトキューにキューイングできなかった".format(detail))
		else:
			print("\t\t0x{:02X}:unkown result".format(detail))


	def wait_for_connect(self,clientAddr):
		"""子機接続完了待ち合わせ(PANコーディネータ向け)
		"""
		while(True):
			recvData = self.recv()
			if(len(recvData) == 0):
				continue

			cmdCode = get_cmd_code(recvData)
			cmdData = get_cmd_data(recvData)
			if(cmdCode == 0x601A): # 接続状態変更通知(0x601A)
				state = cmdData[0]
				mac = cmdData[1:9]
				strMac = "".join(format(x, "02X")for x in mac)
				rssi = cmdData[9] - 256
				if (state == 0x01): # MAC接続完了
					print("Notify MAC connection \'{:s}\'".format(strMac))
				elif (state == 0x02): # PANA認証完了
					print("Notify PANA auth completion \'{:s}\'".format(strMac))
					if( mac == clientAddr): return 0
				elif (state == 0x03): # MAC切断
					print("Notify MAC disconnection \'{:s}\'".format(strMac))
				elif (state == 0x04): # PANA切断
					print("Notify PANA disconnection \'{:s}\'".format(strMac))
				continue
			elif(cmdCode == 0x6023): # HAN 受入れ接続モード変更通知(0x6023)
				print("The module has returned to normal mode.\nPlease restart this script.")
				return 1
			else:
				continue

		return 0

	def wait_for_pana_auth(self):
		"""親機とのPANA認証完了待ち合わせ(コーディネータorエンドデバイス向け)
		"""
		while(True):
			if( self.isSleepingDevice ):
				self.do_cmd_han_poll_req()

			recvData = self.recv()
			if(len(recvData) == 0):
				continue

			cmdCode = get_cmd_code(recvData)
			cmdData = get_cmd_data(recvData)
			if(cmdCode == 0x6028): # PANA認証結果通知(0x6028)
				panaRslt = cmdData[0]
				mac = cmdData[1:9]
				strMac = "".join(format(x, "02X")for x in mac)
				if (panaRslt == 0x01): # PANA認証成功
					print("PANA Auth Complete \'{:s}\'".format(strMac))
					break
				else:
					print("PANA Auth failed(result={:02X}).".format(panaRslt))
					return 1
			elif(cmdCode == 0x2061):
				sleep(0.2)
				continue
			else:
				continue

		# ポールリクエスト応答を待ち合わせる
		if( self.isSleepingDevice ):
			self.wait_for_recv_command(0x2061)

		return 0

	def cmd_han_poll_req(self):
		"""HAN:ポールリクエスト送信(cmd=0x0061->0x2061)
		"""
		self.do_cmd_han_poll_req()
		return self.do_cmd_han_poll_res()

	def do_cmd_han_poll_req(self):
		sendData = build_req_cmd(0x0061)
		return self.send(sendData)

	def do_cmd_han_poll_res(self):
		# response
		while(True):
			recvData = self.recv()
			if(len(recvData) == 0):
				self.do_cmd_han_poll_req()
				continue
			if(get_cmd_code(recvData) == 0x2061):
				break

		cmdData = get_cmd_data(recvData)
		if(cmdData[0] == COMMAND_SUCCESS):
			# print("ポールリクエスト送信: Success")
			return 0, cmdData
		else:
			print("ポールリクエスト送信: Failure")
			return 1, cmdData

	def cmd_com_activescan(self, _time, _chmask, _id, _pairID):
		"""共通:アクティブスキャン実行(cmd=0x0051->0x2051->0x4051)
		"""
		cmdData = bytearray()
		cmdData += _time.to_bytes(1, byteorder='big')
		cmdData += _chmask.to_bytes(4, byteorder='big')
		cmdData += _id.to_bytes(1, byteorder='big')
		if( _id == 0x01):
			cmdData += _pairID.to_bytes(8, byteorder='big')
		sendData = build_req_cmd(0x0051, list(cmdData))
		self.send(sendData)

		# wait notify and response
		results = []
		while(True):
			recvData = self.recv()
			cmdCode = get_cmd_code(recvData)
			cmdData = get_cmd_data(recvData)
			if(cmdCode == 0x2051):
				if(cmdData[0] == COMMAND_SUCCESS):
					print("アクティブスキャン実行: Success")
					return 0, results
				else:
					print("アクティブ実行: Failure")
					return 1, results
			elif(cmdCode == 0x4051):
				reply = True if(cmdData[0] == 0x00) else False
				ch = cmdData[1]
				if(reply):
					numOfScan = cmdData[2]
					print(" スキャン結果通知: ch{:d} 応答あり({:d}個)".format(ch,numOfScan))
					for i in range(numOfScan):
						index = 3+i*11
						addr = cmdData[index:index+8]
						strAddr = "".join(format(x, "02X")for x in addr)
						index += 8
						panid = int.from_bytes(cmdData[index:index+2], byteorder='big')
						index += 2
						rssi = cmdData[index] -256
						print("\tMAC:{:s},\tPAN-ID:{:04X},\tRSSI:{:d}dBm".format(strAddr,panid,rssi))
						results.append([ch, addr, panid, rssi])
				else:
					print(" スキャン結果通知: ch{:d} 応答なし".format(ch))
		
		return 0, results

	def cmd_han_disconnect_mac(self, mac):
		"""HAN:HAN切断(cmd=0x00D3->0x20D3,0x206A)
		"""
		cmdData = bytearray()
		cmdData += mac
		sendData = build_req_cmd(0x00D3, cmdData)
		self.send(sendData)
		# response
		ret_D3 = self.chk_response("HAN切断",0x20D3, COMMAND_SUCCESS)
		ret_6A = self.chk_response("HANデバイスリスト削除",0x206A,
 			COMMAND_SUCCESS)

		ret = 0
		if(ret_D3 != 0) or (ret_6A != 0):
			ret = 1

		return ret

	def cmd_han_get_pana_info(self):
		"""HAN:HAN PANA認証情報取得(cmd=0x002D->0x202D)
		"""
		sendData = build_req_cmd(0x002D)
		self.send(sendData)

		# response
		recvData = self.recv()
		cmdCode = get_cmd_code(recvData)
		cmdData = get_cmd_data(recvData)
		if(cmdCode == 0x202D) and (cmdData[0] == COMMAND_SUCCESS):
			print(" PANA認証情報取得： Success")
			if(self.hanMode == HAN_MODE_PANC) \
				or (self.hanMode == HAN_MODE_DUAL):
				numOfSettings = cmdData[1]

				print(" 設定数： {:02d}".format(numOfSettings))
				idx = 2
				for i in range(numOfSettings):
					# MACアドレス
					mac = cmdData[idx:idx+8]
					strmac = "".join(format(x, "02X")for x in mac)
					idx += 8
					# パスワード
					pwd = cmdData[idx:idx+16]
					strpwd = "".join(format(x, "02X")for x in pwd)
					print("\t{:s}, pass={:s}".format(strmac,strpwd))
					idx += 16
			else:
				# パスワード
				pwd = cmdData[1:1+16]
				strpwd = "".join(format(x, "02X")for x in pwd)
				print("\tpass={:s}".format(strpwd))

			ret = 0
		else:
			print("HAN PANA認証情報取得: Failure")
			ret = 1

		return ret

