#!/usr/bin/python3
# coding:utf-8

# Copyright (c) 2019 ISB CORPORATION <http://www.isb.co.jp/>
# All rights reserved.

import sys
import os
import signal
import json
#import re
import time
from time import sleep

import isb_wsn_serial
from isb_wsn_serial import WiSunSerial

################################################################################
# Wi-SUN Dual Stack
################################################################################

def print_panc_menu():
	print('\nSample Program Menu:')
	print('\t1. Unicast demo')
	print('\t2. Multicast demo')
	print('\tc. Change destination address')
	print('\th. show this')

def panc_menu(ser, ipv6):
	print_panc_menu()
	destIpv6 = ipv6

	input_str = ""
	while input_str == "":
		input_str = input('>>> ')

		# コマンド実行前に受信済みデータを破棄
		ser.flush_recv_buf()

		if input_str == "":
			continue
		elif input_str == "h":
			print_panc_menu()
		elif input_str == "1":
			# Unicast
			payload = bytes([i%256 for i in range(5)])
			ret = ser.cmd_com_send_data(
				destIpv6, 1111,1111, payload)
			if(ret != 0): 
				input_str = ""
				continue

			# receive echoback
			recv_echoback(ser)
		elif input_str == "2":
			# Multicast
			mAddr = "FF020000000000000000000000000001"
			payload = bytes([i%256 for i in range(10)])
			ret = ser.cmd_com_send_data(
				mAddr, 2222,2222, payload)
			if(ret != 0): 
				input_str = ""
				continue

			# receive echoback
			recv_echoback(ser)
		elif input_str == "c":
			# 送信先IPアドレスの変更
			mac = isb_wsn_serial.input_dest_mac()
			destIpv6 = isb_wsn_serial.mac2ipv6(mac)
		else:
			print('unkown')
			print_panc_menu()

		# continue loop
		input_str = ""

def recv_echoback(ser):
	print("wait for receive echoback...")
	errcnt = 0

	while(errcnt < 2):
		recvData = ser.recv()
		cmdCode = isb_wsn_serial.get_cmd_code(recvData)
		cmdData = isb_wsn_serial.get_cmd_data(recvData)

		if(cmdCode == 0x6018): # receipt notification
			srcIpv6 = cmdData[0:16]
			strSrcIpv6 = "".join(format(x, "02X")for x in srcIpv6)
			destPort = int.from_bytes(cmdData[18:20], byteorder='big')
			rssi = int.from_bytes([ cmdData[24] ], 'big', signed=True)
			payload = cmdData[27:]
			print("Received from {:s}, RSSI={:d}, Port={:d}".format(
				strSrcIpv6, rssi, destPort))
			print(" Data={:s}".format(
				"".join(format(x, "02X")for x in payload)))
			break
		else:
			errcnt += 1
			continue

	if(errcnt > 0):
		print(" -> timeout")


def panc_main():
	"""Wi-SUNメイン処理
	"""
	# 設定をjsonから読み込む
	ttyDev, init_settings, panid, han_devices = load_config()
	# シリアルポートオープン
	wsn_ser = WiSunSerial(ttyDev)
	if(wsn_ser.isOpen() == False): return 1

	# ハードウェアリセット
	wsn_ser.cmd_com_reset()
	wsn_ser.cmd_com_get_addr()

	# 初期設定
	wsn_ser.cmd_com_init(init_settings)

	# HAN動作開始
	wsn_ser.cmd_han_start( panid )
	# HAN PANA開始
	wsn_ser.cmd_han_pana_start()
	# HAN PANA認証情報設定
	wait_dev = None
	for hanDev in han_devices:
		wsn_ser.cmd_han_pana_setting( hanDev[0], hanDev[1])
		wait_dev = hanDev[0]

	#初期接続モード
	wsn_ser.cmd_han_switch_connect_mode(0x01)

	# Open UDP port
	wsn_ser.cmd_com_open_port(1111)
	wsn_ser.cmd_com_open_port(2222)

	# wait for connect the last  device in the config.json
	wsn_ser.wait_for_connect(wait_dev)
	ipv6 = isb_wsn_serial.mac2ipv6(wait_dev)

	# PANコーディネータのメニュー起動
	panc_menu(wsn_ser, ipv6)

	return 0



################################################################################
# Application Config
################################################################################
_g_config = {}

def set_app_config(config_dict):
	global _g_config
	_g_config = config_dict.copy()
	return 

def load_config():
	"""設定をjsonファイルから読み込む
	"""
	basedir = os.path.dirname(os.path.abspath(__file__))
	f = open(basedir+"/panc_config.json", 'r')
	jsonObj = json.load(f)
	f.close()
	# グローバル設定を更新
	set_app_config(jsonObj)

	# シリアルポート
	ttyDev = jsonObj["SerialInterface"]["port"]
	# 動作モード
	mode = jsonObj["InitialSetting"]["actionMode"]
	# HAN Sleep機能
	if( jsonObj["InitialSetting"]["sleepSupport"] ):
		sleepSupport = 0x01
	else:
		sleepSupport = 0x00
	# チャネル
	channel = jsonObj["InitialSetting"]["channel"]
	# 送信出力
	txPower = jsonObj["InitialSetting"]["txpower"]
	# PAN-ID
	panid = int(jsonObj["InitialSetting"]["panid"], 16)

	# 接続対象のHANデバイス
	han_devices = []
	for device in jsonObj["HAN"]:
		addr = bytes.fromhex( device["address"] )
		pas = device["password"].encode("ascii")
		han_devices.append( [ addr, pas ] )

	# 初期設定
	init_settings = [
		mode,			# 動作モード(0x01:PANコーディネータ,0x05:Dual)
		sleepSupport,	# HAN Sleep 機能設定(0x00:無効,0x01:有効)
		channel,		# チャネル(0x04～0x11)
		txPower			# 送信電力(0x00:20mW,0x01:10mW,0x02:1mW)
	]

	return ttyDev, init_settings, panid, han_devices


################################################################################

def sigint_handler(signal, frame):
	print('\nterminated by user')
	sys.exit(0)

if __name__ == '__main__':
	signal.signal(signal.SIGINT, sigint_handler)
	ret = panc_main()
	sys.exit(ret)

